#include "ESP8266_PowerSave.h"
#include "pins.h"

ESP8266_PowerSave esp8266(RX_PIN_TO_ESP8266, TX_PIN_TO_ESP8266);

/* -------------------------------------------------------------------
 * Setup des ATmega zur Kommunikation mit dem ESP8266.
 */
void ESP8266_PowerSave::setup(uint8_t power, uint8_t enable, long spd) {
  enabled=0;
  power_pin=power;
  enable_pin=enable;
  speed=spd;

  pinMode(power_pin, OUTPUT);
  pinMode(enable_pin, OUTPUT);
  digitalWrite(power_pin, HIGH);  // ESP ausschalten
}

/* -------------------------------------------------------------------
 * ESP8266 power on und enable
 */
void ESP8266_PowerSave::enable(void) {
  if(!enabled) {
    digitalWrite(power_pin, LOW);     // ESP8266 mit Strom versorgen
    delay(10);
    digitalWrite(enable_pin, HIGH);   // ESP8266 aktivieren
    esp.begin(speed);
    enabled=1;
    esp.setTimeout(1000);
  }
}

/* -------------------------------------------------------------------
 * ESP8266 disable und power off
 */
void ESP8266_PowerSave::disable(void) {
  digitalWrite(tx_pin, LOW);      // keine Spannung am abgeschalteten ESP8266
  digitalWrite(enable_pin, LOW);  // ESP8266 deaktivieren
  delay(10);
  digitalWrite(power_pin, HIGH);  // Strom für ESP8266 abschalten
  enabled=0;
}

/* -------------------------------------------------------------------
 * Stellt den ESP8266 auf
 *    "station mode", 
 *    "single connection" und
 *    connect to AP "FreifunkWees01.2 (http://ffw)"
 * ein.
 */
void ESP8266_PowerSave::init(void) {
  esp.println(F("AT+CWMODE_CUR=1"));
  esp.println(F("AT+CIPMUX=0"));
  esp.println(F("AT+CWAUTOCONN=1"));
  esp.println(F("AT+CWJAP_CUR=\"FreifunkWees01.2 (http://ffw)\",\"\""));
}

/* -------------------------------------------------------------------
 * "connect_retries" mal versuchen, die Verbindung zu "host":"port"
 * herzustellen.
 * Liefert 1, wenn verbunden. Sonst 0.
 */
uint8_t ESP8266_PowerSave::connect_TCPIP(char *host, unsigned int port, uint8_t connect_retries)  {
  for(uint8_t i=0; i<connect_retries; i++) {
    esp.print("AT+CIPSTART=\"TCP\",\"");
    esp.print(host);
    esp.print("\",");
    esp.println(port);
    if(esp.find((char*)"\r\nOK\r\n"))  {
      delay(50);  // den braucht er....
      if(is_connected_via_TCPIP()) {
        return(1);
      }
    }
  }
  return(0);
}

/* -------------------------------------------------------------------
 * Verbindungs-Status abfragen (3 = connected via TCP/IP).
 * Liefert 1, wenn der ESP8266 via TCP/IP verbunden ist. Sonst 0.
 */
uint8_t ESP8266_PowerSave::is_connected_via_TCPIP(void) {
  return(get_TCPIP_status()==3);
}

/* -------------------------------------------------------------------
 * Verbindung zum Python-Script schließen.
 * Liefert 1, wenn die Antwort eine OK-Zeile enthielt. Sonst 0.
 */
char ESP8266_PowerSave::disconnect_TCPIP(void)  {
  for(char i=0; i<3; i++) {
    esp.println(F("AT+CIPCLOSE"));
    if(esp.find((char*)"\r\nOK\r\n"))  {
      // mit dem jetzt voraussichtlich folgenden Chip-Disable
      // und PowerOff noch etwas warten, damit das auch an der
      // Gegenseite ankommt - und dort nicht auf timeout läuft.
      delay(50);
      return(1);
    }
  }
  return(0);
}

/* -------------------------------------------------------------------
 * Schaltet den ESP8266 ein, wartet bis er bereit ist, initialisiert
 * ihn und wartet dann darauf, dass er eine IP-Adresse bekommen hat.
 */
char ESP8266_PowerSave::power_on_and_wait_for_ip(char ip_timeout) {
  enable();
  wait_until_responsive(10);
  if(wait_for_ip(ip_timeout))  {
    return(1);
  }
  return(0);
}

/* -------------------------------------------------------------------
 * Sendet "sendbuf_len" Byte aus "sendbuf" an den via TCP verbundenen
 * ESP8266 mit einer angehängten 2-Byte-Checksum.
 * Die empfangene Antwort wird in "recbuf" abgelegt, sofern die
 * Anzahl der empfangenen Byte kleiner "recbuf_len" ist. Wenn die
 * empfangene Checksum zur aus den empfangenen Daten berechneten
 * Checksum passt, wird die Anzahl der in "recbuf" abgelegten Zeichen
 * zurückgeliefert. Ein negativer Rückgabewert zeigt einen Fehler an.
 */
int ESP8266_PowerSave::TCPIP_send_receive(char *sendbuf, int sendbuf_len, char *recbuf, int recbuf_len) {
  int len, chksum;
  uint8_t rc;

  esp.print(F("AT+CIPSEND="));                      // Sende-Kommando an ESP8266
  esp.println(sendbuf_len+2);                       // Länge der zu sendenden Daten + Checksum
  delay(20);                                        // etwas warten zwischen Kommando und Daten - sonst klappts nicht
  esp.write(sendbuf, sendbuf_len);                  // Daten senden
  writeInt(getBinChecksum(sendbuf, sendbuf_len));   // 16 Bit Checksum senden
  // Antwort: "+IPD,nn:<nn Byte Daten>" (bei CIPMUX=0 = "single connection")
  if(!esp.find((char*)"+IPD,")) {                   // wartet gemäß Stream.timeout (Default=1Sek) auf Daten
    return(-1);                                     // keine Antwort erhalten
  }
  len=esp.parseInt()-2;                             // Länge(ASCII) als integer lesen, -2 für Checksum
  if(len>=recbuf_len) {
    emptyBuffer();
    return(-2);                                     // da kommen mehr Daten, als in recbuf passen
  }
  esp.find(':');
  esp.readBytes(recbuf, len);                       // Daten lesen ohne Checksum
  esp.readBytes((char*)&chksum, 2);                 // 2 Byte Checksum lesen
  if(chksum==getBinChecksum(recbuf, len))  {        // Checksum gegen Checksum der empfangenen Daten prüfen
    rc=len;                                         // Anzahl empfangener Zeichen zurückliefern
  } else {
    rc=-3;                                          // Checksum der Antwort ist falsch
  }
  delay(20);                                        // dem Empfänger (oder ESP8266) etwas Zeit zur Verarbeitung geben
  return(rc);
}

/* -------------------------------------------------------------------
 *  Sendet die Null-terminierte Zeichenfolge "sendbuf" und liefert die 
 *  Antwort als Null-terminierte Zeichenfolge in "recbuf" mit der
 *  Maximallänge "recbuf_len" zurück.
 */
int ESP8266_PowerSave::TCPIP_String_send_receive(char *sendbuf, char *recbuf, int recbuf_len) {
  int lng;

  esp.setTimeout(2000);   // der GET_TIME kann auf Python-Seite bis zu einer Sekunde dauern (wegen "precise=True")
  lng=TCPIP_send_receive(sendbuf, strlen(sendbuf), recbuf, recbuf_len-1);
  esp.setTimeout(1000);

  if(lng>0) {
    *(recbuf+lng)='\0';
    return(1);
  }
  return(0);
}




/* -------------------------------------------------------------------
 * Receive-Buffer leerlesen (und ggf. vorhandenen Inhalt anzeigen).
 */
void ESP8266_PowerSave::emptyBuffer(void)  {
  char c;
  unsigned long timeout;

  if(esp.available()==0)  {
    return;
  }
  timeout=millis()+100;  // maximal 1/10 Sekunde lang leerlesen
  while(esp.available() && millis()<timeout)  {
    c=esp.read();
  }
}

/* -------------------------------------------------------------------
 * Wartet (bis zu "timeout_seconds" Sekunden) darauf, dass der ESP8266
 * mit "OK" auf einen "AT"-Befehl reagiert.
 * Liefert 1, wenn der ESP8266 korrekt reagiert hat. Sonst 0.
 */
char ESP8266_PowerSave::wait_until_responsive(uint8_t timeout_seconds)  {
  for(uint8_t i=0; i<timeout_seconds; i++)  {
    esp.println(F("AT"));
    if(esp.find((char*)"\r\nOK\r\n"))  { // wartet den Default-Timeout von einer Sekunde
      emptyBuffer();
      return(1);
    }
  }
  return(0);
}

/* -------------------------------------------------------------------
 * Liefert den Status des ESP8266 als Zahl.
 *  0 : Fehler
 *  2 : Got IP
 *  3 : Connected     (via  TCP/IP)
 *  4 : Disconnected  (from TCP/IP)
 */
uint8_t ESP8266_PowerSave::get_TCPIP_status(void) {
  uint8_t state=0;
  esp.println(F("AT+CIPSTATUS"));
  if(esp.find((char*)"\r\nSTATUS:")) {
    state=esp.parseInt();
  }
  esp.find((char*)"\r\nOK\r\n"); // Rest weglesen
  return(state);
}

/* -------------------------------------------------------------------
 * Wartet maximal "timeout_seconds" darauf, bis der ESP8266 nach
 * Chip-Enable eine IP-Adresse bekommt.
 * Liefert 1, wenn der ESP8266 eine IP-Adresse hat. Sonst 0.
 */
char ESP8266_PowerSave::wait_for_ip(char timeout_seconds)  {
  for(char i=0; i<timeout_seconds*10; i++)  {
    if(get_TCPIP_status()==2) {
      return(1);
    }
    delay(100);
  }
  return(0);
}

/* -------------------------------------------------------------------
 *  Sendet "d" an den via TCP verbundenen ESP8266 - das niederwertige 
 *  Byte zuerst.
 */
void ESP8266_PowerSave::writeInt(unsigned int d) {
  esp.write((char*)&d, 2);
}

/* -------------------------------------------------------------------
 *  
 */
unsigned int ESP8266_PowerSave::getBinChecksum(char *buf, int len) {
  unsigned int cs=0;
  for(; len>0; len--)  {
    cs+=(unsigned char)*buf++;
  }
  return(cs);
}


