/* *******************************************************************
 *  Temperatur- und Helligkeits-Logger
 *  
 *  Detlev Ahlgrimm, 11.2016 - 01.2017
 *  
 *  21.01.2017  v1.4  Eintritt in den Stundenwechsel verzögert, um zu kurze deepSleep-Zeiten zu kompensieren
 */

#include "DS18B20_temp.h"
#include "RTC_EEPROM.h"
#include "deep_sleep.h"
#include "mem_mgnt.h"
#include "ESP8266_PowerSave.h"
#include "globals.h"
#include "pins.h"

// in globals.h befinden sich noch weitere define-Statements!
#define HOSTNAME "Logger-01"
#define FIRMWARE "1.4"

#define SOCKETSERVER_IP   "192.168.42.80" // ss
//#define SOCKETSERVER_IP   "192.168.42.99" // ffw
#define SOCKETSERVER_PORT 2627

#define VOLTAGE_FIX   -0.3          // Korrektur-Wert für die Spannung an VOLTAGE_READING_PIN

#define SLEEP_MS  4864              // Anzahl Millisekunden für eine Tiefschlaf-Phase (4096+512+256)
#define SLEEPS    60000/SLEEP_MS-1  // Anzahl Tiefschlaf-Phasen für eine volle Minute
// der -1 hat den Sinn, am Ende der Minute etwas mehr Zeit zur Verarbeitung zu haben

#define SYNC_HOUR  19               // täglich um 19:00:00 Uhr werden die Daten ans Python-Script gesendet

#define SYNC_HOUR1  7               // Sync-Zeiten bei zwei Sensor-Paaren
#define SYNC_HOUR2 19               // immer +12 (oder -12) von SYNC_HOUR1

byte set_timestamp_at_next_minute;  // ein Flag

char error;                         // zur Steuerung der Blink-Frequenz der Lebenszeichen-LED
// Bit 0 : letzter Sync ist fehlgeschlagen
// Bit 1 : Batterie der RTC scheint leer zu sein



/* -------------------------------------------------------------------
 *  Initialisierung
 */
void setup() {
  unsigned long sc;     // Restzeit bis zur nächsten Minute
  RTCdata t;            // Uhrzeit-Speicher
  byte time_set;        // Flag, dass die GET_TIME-Loop verlassen werden kann

  #if DEBUG==1
  Serial.begin(115200);
  Serial.println("---setup()---");
  #endif

  error=0;
  warnings_g=1;         // Restart vermerken

  pinMode(ALIVE_LED_PIN, OUTPUT);
  pinMode(LDR_POWER_PIN, OUTPUT);
  digitalWrite(ALIVE_LED_PIN, LOW); // LED aus
  digitalWrite(LDR_POWER_PIN, LOW); // LDR_POWER aus

  pinMode(SPECIAL_SYNC_BUTTON, INPUT);    // kein INPUT_PULLUP - der zieht zuviel Strom
  digitalWrite(SPECIAL_SYNC_BUTTON, LOW); // der Pullup wird nur bei Bedarf zugeschaltet

  RTC_EEPROM_init();            // Ports für die Kommunikation mit dem RTC/EEPROM-Modul setzen
  discoverOneWireDevices();     // erkennt die DS18B20-Sensoren und setzt "sensor_count_g"
  esp8266.setup(ESP8266_POWER_PIN, ESP8266_CHIP_ENABLE_PIN, 38400); // Ports für ESP8266 einstellen

  for(byte i=0; i<sensor_count_g; i++) {
    readDS18B20_CRC(i);   // Dummy-Read
  }

  if(specialSyncButtonPressed()) {
    // Standort-Wechsel-Wunsch
    digitalWrite(ALIVE_LED_PIN, HIGH); // LED ein
    invalidateSensorHours();
    warnings_g|=2;      // vermerken, dass das Python-Script neue Standort-Sätze anlegen soll
  }

  time_set=0;
  for(int i=0; i<10; i++) {     // 10 Versuche, die Zeit vom Python-Script zu holen
    if(esp8266.power_on_and_wait_for_ip(10)) {
      if(esp8266.connect_TCPIP((char*)SOCKETSERVER_IP, SOCKETSERVER_PORT, 10)) {
        if(esp8266.TCPIP_String_send_receive((char*)"GET_TIME", buffer_g, BUFFER_SIZE)) {
          #if DEBUG==1
          Serial.println(buffer_g);
          #endif
          digitalWrite(RTC_EEPROM_POWER_PIN, HIGH);
          setRTC(buffer_g);         // Uhr stellen gemäß buffer_g, Format: YYYY.MM.DD hh:mm:ss
          digitalWrite(RTC_EEPROM_POWER_PIN, LOW);
          time_set=1;
        }
        esp8266.disconnect_TCPIP();
      }
    }
    esp8266.disable();
    if(time_set==1) {
      break;
    }
    deepSleep(8192);    // hat nicht geklappt -> acht Sekunden Tiefschlaf vor dem nächsten Versuch
  }
  digitalWrite(ALIVE_LED_PIN, LOW); // LED aus

  if(time_set==0) {
    error|=1;     // Kennung für "Uhrzeit konnte nicht geholt werden" setzen
  }

  clearSensorHour();                        // den ein-Stunden-Speicher mit Nullen überschreiben (nur Optik)
  sensorRAM_idx_g=0;                        // Index des ein-Minuten-Speichers
  t=getRTC();                               // Zeit holen
  if(t.year==0) {                           // Batterie der RTC leer?
    warnings_g|=4;
    error|=2;
  }
  power_on_timestamp_g=t;                   // Power-On-Zeit merken
  if((60-t.second)*1000l<SLEEP_MS) {        // wenn Restzeit bis zur nächsten Minute kleiner Tiefschlaf-Dauer
    deepSleep(2*SLEEP_MS);                  // zweimal Tiefschlaf, damit er "sauber" in loop() eintritt
    t=getRTC();
  }
  for(byte i=0; i<sensor_count_g; i++) {
    sensorHourRAM_g[i].timestamp=t;         // Zeit im ein-Stunden-Speicher merken
  }
  set_timestamp_at_next_minute=false;       // mit "kein Ende der Stunde" initialisieren
  sensorHourRAM_idx_g=t.minute;             // Start-Index für den ein-Stunden-Speicher setzen (0-59)
  sc=(60-t.second)*1000l;                   // Millisekunden bis zur nächsten Minute (>=1000 & <=60000)
  cycles_for_full_minute_g=sc/SLEEP_MS-1;   // Anzahl der Tiefschlaf-Phasen bis zur nächsten Minute (>=-1 & <=11)
  minutes_to_full_hour_g=60-t.minute;       // Minuten bis zur nächsten Stunde (>=1 & <=60)
  #if DEBUG==1
  Serial.print("sensor_count_g=");
  Serial.println(sensor_count_g);
  delay(10);
  #endif
}



/* -------------------------------------------------------------------
 *  do forever
 */
void loop() {
  static byte rc;                 // allgemeiner Rückgabewert
  static byte i;                  // allgemeine Schleifenvariable
  static int retry_hour=-1;       // wenn zwischen 0 und 23 -> alle 5 Minuten einen Sendeversuch
  static long sc;                 // Restzeit bis zur nächsten Minute
  static RTCdata t, t2;           // Uhrzeit-Speicher
  static SensorData d;            // ein Messwerte-Paar
  
  processSensorData();                      // Sensor-Daten holen, in sensorRAM_g[][] speichern und sensorRAM_idx_g inkrementieren

  deepSleep(SLEEP_MS-512);                  // eine Tiefschlaf-Phase schlafen
  blinkAliveLED(error);                     // schläft nochmal 512ms

  if(specialSyncButtonPressed()) {          // wenn Sonder-Sync-Taster gedrückt ist
    #if DEBUG==1
    Serial.println("\nSonder-Sync");
    #endif
    rc=sendAllHours();            // alles senden, was bereits im EEPROM ist
    if(rc==1) {
      error&=0xfe;
    } else {
      // schlägt der Sonder-Sync fehl, muss er per Taster wiederholt werden
      error|=1;   // aber trotzdem das Flag für die LED setzen, dass der Sync fehlgeschlagen ist
    }
    // das Ende der Minute bzw. "cycles_for_full_minute_g" muss jetzt korrigiert werden
    t=getRTC();
    if(t.minute!=sensorHourRAM_idx_g) {       // wenn die neue Minute schon angefangen hat
      cycles_for_full_minute_g=-1;            // Minuten-Ende sofort betreten und darin keine weiteren Sensor-Daten mehr holen
    } else {
      sc=(60-t.second)*1000l;                 // Millisekunden bis zur nächsten Minute
      cycles_for_full_minute_g=sc/SLEEP_MS;   // Anzahl der Tiefschlaf-Phasen bis zur nächsten Minute neu setzen
    }
    #if DEBUG==1
    Serial.print("t.minute=");
    Serial.print(t.minute, DEC);
    Serial.print("  sensorHourRAM_idx_g=");
    Serial.print(sensorHourRAM_idx_g);
    Serial.print("  sc=");
    Serial.print(sc, DEC);
    Serial.print("  sc/SLEEP_MS=");
    Serial.print(sc/SLEEP_MS, DEC);
    Serial.print("  cycles_for_full_minute_g=");
    Serial.println(cycles_for_full_minute_g, DEC);
    #endif
  }

  if(--cycles_for_full_minute_g<=0) {       // wenn das Ende der Minute nah oder erreicht ist
    /* --------------------------------------------------------------------------------------------------- */
    if(cycles_for_full_minute_g>=0) {       // nur dann, wenn die Minute noch nicht zuende ist...
      processSensorData();                  // ...nochmal Sensor-Daten holen, in sensorRAM_g[][] speichern und sensorRAM_idx_g inkrementieren
    }
    for(i=0; i<sensor_count_g; i++) {
      d=getSensorDataAvg(i);                          // Mittelwerte der Sensor-Daten im ein-Minuten-Speicher berechnen
      sensorHourRAM_g[i].data[sensorHourRAM_idx_g]=d; // Sensor-Daten in den ein-Stunden-Speicher kopieren
    }

    #if DEBUG==1
    if(retry_hour>=0) {
      Serial.print("\nretry_hour="); Serial.print(retry_hour); Serial.print("  sensorHourRAM_g[0].timestamp.hour=");  Serial.print(sensorHourRAM_g[0].timestamp.hour, DEC);
      Serial.print("  sensorHourRAM_idx_g="); Serial.println(sensorHourRAM_idx_g, DEC);
    }
    #endif

    if(retry_hour==sensorHourRAM_g[0].timestamp.hour && (sensorHourRAM_idx_g%5)==0) { // wenn letzter Sync fehlgeschlagen ist, alle 5 Minuten wiederholen
      rc=sendAllHours();  // alles Senden
      if(rc==1) {
        retry_hour=-1;    // hat diesmal geklappt -> nicht länger wiederholen
        error&=0xfe;
      }
      // jetzt hat die neue Minute voraussichtlich schon angefangen
      // der folgende "cycles_for_full_minute_g=..." wird ggf. unten korrigiert
    }
    cycles_for_full_minute_g=SLEEPS;          // Anzahl Tiefschlaf-Phasen für die neue Minute einstellen

    #if DEBUG==1
    Serial.print("\nsensorRAM_idx_g=");
    Serial.print(sensorRAM_idx_g, DEC);       // Anzahl Werte im ein-Minuten-Speicher
    Serial.print("  sensorHourRAM_idx_g=");
    Serial.print(sensorHourRAM_idx_g, DEC);   // aktueller Index im ein-Stunden-Speicher
    Serial.print("  ");
    printSensorData(d);                       // Mittelwert der Messwerte der Minute anzeigen
    #endif

    sensorHourRAM_idx_g++;                // Index im ein-Stunden-Speicher auf nächste Minute einstellen
    sensorRAM_idx_g=0;                    // Index im ein-Minuten-Speicher für nächste Minute einstellen

    if(--minutes_to_full_hour_g<=0) {     // wenn das Ende der Stunde erreicht ist
      /* --------------------------------------------------------------------------------------------------- */
      for(i=0; i<sensor_count_g; i++) {
        writeSensorHourToEEPROM(i);       // den ein-Stunden-Speicher ins EEPROM kopieren
      }
      /*#if DEBUG==1
      Serial.println();
      Serial.print("checksum[0]="); Serial.println(sensorHourRAM_g[0].checksum, HEX);
      Serial.print("checksum[1]="); Serial.println(sensorHourRAM_g[1].checksum, HEX);
      Serial.println();
      print_sensorHourRAM(0);             // Ausgabe des ein-Stunden-Speichers
      Serial.println("\n"); dumpEEPROM();
      Serial.println();     dumpATmegaEEPROM();
      #endif*/

      set_timestamp_at_next_minute=true;  // ab neuer Minute die aktuelle Zeit im ein-Stunden-Speicher setzen

      if((sensor_count_g==2 && (sensorHourRAM_g[0].timestamp.hour+1==SYNC_HOUR1 || sensorHourRAM_g[0].timestamp.hour+1==SYNC_HOUR2)) ||
         (sensor_count_g==1 &&  sensorHourRAM_g[0].timestamp.hour+1==SYNC_HOUR) ) {   // vorgesehene Sync-Zeit am Ende dieser Minute erreicht
        #if DEBUG==1
        Serial.println("Sync...");
        #endif
        rc=sendAllHours();
        if(rc!=1) {
          #if DEBUG==1
          Serial.println("fehler !!!!");
          #endif
          retry_hour=sensorHourRAM_g[0].timestamp.hour+1; // alle fünf Minuten in der folgenden Stunde einen neuen Sendeversuch starten
          error|=1;
        } else {
          error&=0xfe;
        }
        // Der Sync kann gerne etwas länger dauern....und auch erst in der nächsten Minute
        // fertig werden. In diesem Fall werden lediglich entsprechend weniger Messwerte
        // für die erste Minute der nächsten Stunde erfasst.
        // Es darf jedoch nicht länger als eine Minute sein - sonst fehlen nachher Minuten-Werte.
        // Ausserdem muss ein Sende-Fehlschlag berücksichtigt werden. In diesem Fall wäre
        // nach 5 Minuten ein weiterer Sende-Versuch einzuleiten.
      }
    }

    // sicher in die neue Minute eintreten
    t=getRTC();
    sc=(60-t.second)*1000l;                   // Millisekunden bis zur nächsten Minute
    if(t.minute==(sensorHourRAM_idx_g%60)) {  // wenn die neue Minute schon angefangen hat (der %60 macht aus 60 -> 0)
      cycles_for_full_minute_g=sc/SLEEP_MS-1; // Anzahl der Tiefschlaf-Phasen bis zur nächsten Minute neu setzen
    } else {                                  // wenn noch alte Minute
      while(sc>SLEEP_MS) {                    // bis zur nächsten Minute schlafen
        deepSleep(SLEEP_MS-512);              // eine Tiefschlaf-Phase schlafen
        blinkAliveLED(4);                     // schläft nochmal 512ms
        sc-=SLEEP_MS;
      }
      deepSleep(sc+16);                       // restliche Zeit kleiner einer Tiefschlaf-Phase warten
      // Die +16 deswegen, weil deepSleep() bei Zeiten <16ms nicht wartet, die neue Minute aber
      // sicher angefangen haben soll.
      // Wurde "sc" z.B. eine Millisekunde vor der neuen Minute ermittelt, beträgt die Wartezeit
      // 1016ms. Somit ist die erste Sekunde der neuen Minute zwar bereits abgelaufen - aber das
      // stört nicht.
      // Andersrum, wenn "sc" eine Millisekunde nach Sekunden-Wechsel ermittelt würde, und
      // sc==1000 wäre, müsste deepSleep() mindestens 999 Millisekunden warten, damit die neue
      // Minute erreicht wäre. deepsleep() würde jedoch warten für: 512+256+128+64+32=992ms
      // Die neue Minute hätte hier also noch nicht angefangen.....
    }

    // ggf. in die neue Stunde eintreten
    if(set_timestamp_at_next_minute==true) {  // wenn eben das Ende einer Stunde verarbeitet wurde
      set_timestamp_at_next_minute=false;     // Kenner dafür zurücksetzen
      deepSleep(SLEEP_MS-512);                // v1.4: vorsichtshalber noch eine Tiefschlaf-Phase schlafen
      blinkAliveLED(4);                       // schläft nochmal 512ms
      clearSensorHour();                      // RAM-Speicher sauber machen
      t=getRTC();                             // aktuelle Zeit holen
      for(i=0; i<sensor_count_g; i++) {
        sensorHourRAM_g[i].timestamp=t;       // aktuelle Zeit im ein-Stunden-Speicher setzen
      }
      sc=(60-t.second)*1000l;                 // v1.4: Millisekunden bis zur nächsten Minute
      cycles_for_full_minute_g=sc/SLEEP_MS-1; // v1.4: Anzahl der Tiefschlaf-Phasen bis zur nächsten Minute neu setzen
      minutes_to_full_hour_g=60-t.minute;     // Minuten bis zur nächsten Stunde einstellen
      sensorHourRAM_idx_g=t.minute;           // Start-Index für den ein-Stunden-Speicher setzen (0-59)
    }

    #if DEBUG==1
    t2=getRTC();
    Serial.print("\nminutes_to_full_hour_g=");
    Serial.print(minutes_to_full_hour_g);
    Serial.print("  t=");  
    printRTC(t); 
    Serial.print("  t2=");  
    printRTC(t2); 
    Serial.println("\n");
    delay(10);
    #endif
  }
}


/* -------------------------------------------------------------------
 *  Sendet alle Messwerte zum Python-Script und setzt danach die RTC
 *  gemäß Zeit vom Python-Script neu.
 *  Liefert 1, wenn alle Daten erfolgreich gesendet wurden. Sonst 0.
 */
byte sendAllHours(void) {
  static byte sensor_number;    // zum toggle'n zwischen Sensor-Paaren
  static byte rc;               // allgemeiner Rückgabewert
  static byte all_good;         // wird auf 0 gesetzt, wenn mindestens ein Block nicht gesendet werden konnte
  static byte h;                // Schleifenvariable, Stunde
  static byte s;                // Schleifenvariable, Sensor-Paar
  static byte retries;          // Schleifenvariable für mehrfache Übertragungsversuche
  static byte flg;              // ein Flag (==1, wenn erfolgreiche gesendet wurde)
  RTCdata t;                    // Uhrzeit-Speicher
  float voltage;                // gemessene Akku-Spannung
  byte  i;                      // allgemeine Schleifenvariable

  restore_ADC();
  analogReference(INTERNAL);
  voltage=analogRead(VOLTAGE_READING_PIN); // Dummy-Read (der erste Wert stimmt IMMER nicht)

  digitalWrite(ALIVE_LED_PIN, HIGH);  // Lebenszeichen-LED einschalten
  all_good=1;
  if(esp8266.power_on_and_wait_for_ip(10)) {
    digitalWrite(ALIVE_LED_PIN, LOW);  // Lebenszeichen-LED ausschalten
    if(esp8266.connect_TCPIP((char*)SOCKETSERVER_IP, SOCKETSERVER_PORT, 10)) {

      // hier wird die Hochfahrzeit des ESP8266 gleichzeitig für den ADC genutzt
      // https://www.arduino.cc/en/Reference/AnalogReference
      // -> After changing the analog reference, the first few readings from analogRead() may not be accurate.
      for(i=0; i<3; i++) {
        voltage=analogRead(VOLTAGE_READING_PIN)*(1.1/1024.0)*5.0;
      }
      analogReference(DEFAULT);
      voltage+=VOLTAGE_FIX;

      // Datum/Zeit mit Python-Script synchronisieren
      if(esp8266.TCPIP_String_send_receive((char*)"GET_TIME", buffer_g, BUFFER_SIZE)) {
        #if DEBUG==1
        Serial.print("Zeit empfangen : ");
        Serial.println(buffer_g);
        #endif
        digitalWrite(RTC_EEPROM_POWER_PIN, HIGH);
        setRTC(buffer_g);         // Uhr stellen gemäß buffer_g, Format: YYYY.MM.DD hh:mm:ss
        digitalWrite(RTC_EEPROM_POWER_PIN, LOW);
        t=getRTC();                               // Zeit holen
        if(t.year==0) {                           // Batterie der RTC leer?
          warnings_g|=4;
          error|=2;
        }
      }

      // Identifizierungs-Daten senden
      snprintf(buffer_g, 100, "IDENT=%s,%s,%d,%.4d.%.2d.%.2d %.2d:%.2d:%.2d,%d,%1d.%02d", HOSTNAME, FIRMWARE, sensor_count_g, 
        (int)power_on_timestamp_g.year+2000, (int)power_on_timestamp_g.month, (int)power_on_timestamp_g.day, 
        (int)power_on_timestamp_g.hour, (int)power_on_timestamp_g.minute, (int)power_on_timestamp_g.second,
        warnings_g, int(voltage), int((voltage-int(voltage))*100));
      for(flg=retries=0; retries<3; retries++) {
        rc=esp8266.TCPIP_send_receive(buffer_g, strlen(buffer_g), buffer_g+100, BUFFER_SIZE-100);
        if(rc>0 && strcmp(buffer_g+100, "OK")==0) {
          flg=1;
          warnings_g=0;   // wenn erfolgreich gemeldet wurde, ist erstmal wieder alles gut
          break;
        }
      }
      if(flg==0) {  all_good=0; } // das Senden hat "retries" mal nicht geklappt

      // Sensor-Daten aus dem EEPROM senden
      memcpy(buffer_g, "SHOUR.1=", 8);  // auf buffer[6] liegt die Sensor-Nummer
      sensor_number=0;
      for(h=0; h<24/sensor_count_g; h++) {  // bis <12 bei zwei Sensor-Paaren, bis <24 bei einem
        digitalWrite(ALIVE_LED_PIN, h&1);   // Lebenszeichen-LED blinken lassen
        for(s=0; s<sensor_count_g && all_good==1; s++) {
          for(flg=retries=0; retries<3; retries++) {
            if(!readHourFromEEPROM(h, s, (byte*)buffer_g+8)) {
              flg=1;  // nicht senden ist immer erfolgreich
              break;  // Checksum hat nicht gepasst
            }
            buffer_g[6]=s+'1';
            rc=esp8266.TCPIP_send_receive(buffer_g, 0xBA+8, buffer_g+8, BUFFER_SIZE-8);
            if(rc>0 && strcmp(buffer_g+8, "OK")==0) {
              flg=1;
              break;
            }
          }
          if(flg==0) {  all_good=0; } // das Senden hat "retries" mal nicht geklappt
        }
      }

      // Sensor-Daten aus dem RAM senden
      if(minutes_to_full_hour_g>0) {
        for(s=0; s<sensor_count_g; s++) {
          sensorHourRAM_g[s].checksum=getChecksum((byte*)&sensorHourRAM_g[s], s); // Checksum in sensorHourRAM_g[s] setzen
          memcpy(buffer_g+8, (char*)&sensorHourRAM_g[s], 186);
          buffer_g[6]=s+'1';
          for(flg=retries=0; retries<3; retries++) {
            rc=esp8266.TCPIP_send_receive(buffer_g, 0xBA+8, buffer_g+8, BUFFER_SIZE-8);
            if(rc>0 && strcmp(buffer_g+8, "OK")==0) {
              flg=1;
              break;
            }
          }
          if(flg==0) {  all_good=0; } // das Senden hat "retries" mal nicht geklappt
        }
      }
      esp8266.disconnect_TCPIP();
    } else {                          // connect_TCPIP() fehlgeschlagen
      all_good=0;
    }
  } else {                            // power_on_and_wait_for_ip() fehlgeschlagen
    all_good=0;
  }
  esp8266.disable();
  analogReference(DEFAULT);
  digitalWrite(ALIVE_LED_PIN, LOW);   // Lebenszeichen-LED ausschalten
  return(all_good);
}


/* -------------------------------------------------------------------
 *  Liefert True, wenn der Sonder-Sync-Taster gedrückt ist.
 */
byte specialSyncButtonPressed(void) {
  byte rc;
  digitalWrite(SPECIAL_SYNC_BUTTON, HIGH);  // Pullup zur Abfrage des Tasters kurz einschalten
  rc=digitalRead(SPECIAL_SYNC_BUTTON);
  digitalWrite(SPECIAL_SYNC_BUTTON, LOW);   // Pullup wieder aus
  return(rc==0);
}

/* -------------------------------------------------------------------
 *  Überschreibt "sensorHourRAM_g" mit Nullen (nur Optik).
 */
void clearSensorHour(void) {
  for(byte s=0; s<sensor_count_g; s++) {
    sensorHourRAM_g[s].timestamp.year=0;
    sensorHourRAM_g[s].timestamp.month=0;
    sensorHourRAM_g[s].timestamp.day=0;
    sensorHourRAM_g[s].timestamp.hour=0;
    sensorHourRAM_g[s].timestamp.minute=0;
    sensorHourRAM_g[s].timestamp.second=0;
    sensorHourRAM_g[s].checksum=0;
    for(int i=0; i<60; i++) {
      sensorHourRAM_g[s].data[i].temp=0;
      sensorHourRAM_g[s].data[i].ldr=0;
      sensorHourRAM_g[s].data[i].bad=1;
    }
    for(int i=0; i<6; i++) {
      sensorHourRAM_g[s].EEPROM_alignment[i]='\0';
    }
  }
}

/* -------------------------------------------------------------------
 *  Lässt die Lebenszeichen-LED abhängig von "error" kurz aufblinken.
 *  Die Funktion läuft bzw. schläft immer 2*256=512ms.
 */
void blinkAliveLED(byte error) {
  #if DEBUG==1
  Serial.print("blinkAliveLED=");
  Serial.println(error, DEC); // die LED des Arduino UNO ist mit 0.1ms Leuchtdauer nicht zu erkennen
  delay(512);
  return;
  #endif
  if(error==0) {                      // Normales Lebenszeichen -> 1xblinken
    singleBlinkAliveLED();
    sleep_PWR_DOWN(0b000101); // 512ms
  } else if((error&2)==2) {           // die Batterie der RTC ist hinüber -> 3xblinken
    singleBlinkAliveLED();
    sleep_PWR_DOWN(0b000100); // 256ms
    singleBlinkAliveLED();
    sleep_PWR_DOWN(0b000100); // 256ms
    singleBlinkAliveLED();
  } else if((error&1)==1) {           // letzter Sync war nicht erfolgreich -> 2xblinken
    singleBlinkAliveLED();
    sleep_PWR_DOWN(0b000100); // 256ms
    singleBlinkAliveLED();
    sleep_PWR_DOWN(0b000100); // 256ms
  } else if((error&4)==4) {           // Minuten-Wechsel
    singleBlinkAliveLED();
    sleep_PWR_DOWN(0b000101); // 512ms
    singleBlinkAliveLED();
  }
}
void singleBlinkAliveLED(void) {
  digitalWrite(ALIVE_LED_PIN, HIGH);
  delayMicroseconds(100);
  digitalWrite(ALIVE_LED_PIN, LOW);
}

/* -------------------------------------------------------------------
 *  Holt die Sensor-Daten und legt sie im ein-Minuten-Speicher ab.
 */
void processSensorData(void) {
  restore_ADC();                        // erstmal AD-Wandler wieder aktivieren
  digitalWrite(LDR_POWER_PIN, HIGH);    // LDR_POWER ein
  for(byte i=0; i<sensor_count_g; i++) {
    sensorRAM_g[i][sensorRAM_idx_g]=getSensorData(i); // Sensor-Daten im ein-Minuten-Speicher ablegen
  }
  sensorRAM_idx_g++;                    // Index für nächsten Wert einstellen
  digitalWrite(LDR_POWER_PIN, LOW);     // LDR_POWER aus
}


/* -------------------------------------------------------------------
 *  Liefert die Mittelwerte der gesammelten Sensor-Daten aus dem
 *  ein-Minuten-Speicher.
 */
SensorData getSensorDataAvg(byte s) {
  static byte i;
  static byte good_cnt;
  static long temp, ldr;
  static SensorData sd;

  for(temp=ldr=0, good_cnt=i=0; i<sensorRAM_idx_g; i++) {   // sensorRAM_idx_g muss <16 sein!!!
    if(sensorRAM_g[s][i].bad==0) {
      temp+=sensorRAM_g[s][i].temp;   // bis zu 16 mal 12 Bit
      ldr+= sensorRAM_g[s][i].ldr;    // bis zu 64 mal 10 Bit
      good_cnt++;
    }
  }
  if(good_cnt>0) {
    sd.temp=temp/good_cnt;
    sd.ldr= ldr/good_cnt;
    sd.bad=0;
  } else {
    sd.temp=0;
    sd.ldr=0;
    sd.bad=1;
  }
  return(sd);
}


/* -------------------------------------------------------------------
 *  Liefert die gepackten Daten des Sensor-Paares mit der Nummer "nr".
 */
SensorData getSensorData(char nr) {
  static SensorData d;
  int temp;

  if((temp=readDS18B20_CRC(nr))==0x8FFF) {
    d.bad=1;
    return(d);
  }
  d.temp=temp;
  d.ldr=analogRead(LDR_PIN[nr])&0x03FF;
  d.bad=0;
  #if DEBUG==1
  Serial.print("minutes_to_full_hour_g=");  Serial.print(minutes_to_full_hour_g);
  Serial.print("  cycles_for_full_minute_g=");  Serial.print(cycles_for_full_minute_g, DEC);
  Serial.print("  temp="); Serial.print(d.temp/16.0, 2); Serial.print("  ldr="); Serial.print(d.ldr, DEC);  Serial.print("  ");
  delay(10);
  #endif
  return(d);
}


/* -------------------------------------------------------------------
 *  Debug-Print eines SensorData-Elements "d".
 */
#if DEBUG==1
void printSensorData(SensorData d) {
  Serial.print("d.temp=");
  Serial.print(d.temp, DEC);
  Serial.print("  d.temp/16.0=");
  Serial.print(d.temp/16.0, 2);
  Serial.print("  d.ldr=");
  Serial.print(d.ldr, DEC);
  Serial.print("  ");
}
#endif

/* -------------------------------------------------------------------
 *  Debug-Print der Daten einer Stunde.
 */
#if DEBUG==1
void print_sensorHourRAM(byte s) {
  Serial.println();
  printRTC(sensorHourRAM_g[s].timestamp); Serial.println();
  for(byte i=sensorHourRAM_g[s].timestamp.minute; i<60; i++) {
    printSensorData(sensorHourRAM_g[s].data[i]);
    Serial.println();
  }
}
#endif
